/*
 * Copyright (C) 2011 Greg Dorfuss - mhspot.com
 * 
 * SipToSis is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SipToSis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this source code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Author(s):
 * Greg Dorfuss
 */

package local.ua;

import java.util.Vector;

import org.apache.log4j.Logger;

import com.skype.Call;

public class SkypeDTMFSender extends Thread
{
	private boolean stopSender=false;
	private SSCallChannel callChan=null;
	private int digitDelay=0;
    private Vector<Integer> dtmfQueue=new Vector<Integer>();
    private int dtmfQueueSize=0;
    private Logger log = null;

	
	SkypeDTMFSender(SSCallChannel argUa,int argDigitDelay) 
	{
		this.setName(this.getClass().getName()+".T"+this.getName().replaceAll("Thread-", ""));
		log = Logger.getLogger(this.getName());
	    this.callChan=argUa;
	    this.digitDelay=argDigitDelay;
	    this.start();
	}

	
	public void stopDTMFSender()
	{
		this.stopSender=true;
	}
	
	public void run() 
	{
		try
		{
			while (!interrupted())
			{	
				if (stopSender)
					return;
				
				   int fixDigit=getNextDtmfDigit();
				   if (fixDigit>=0)
				   {  
					   for (int c=0;c<this.callChan.skypeChannels.size();c++)
					   {
						   Call curCall=this.callChan.skypeChannels.get(c).getSkypeCall();
						   Call.Status callStat=null;
						   try
						   {
							   callStat=curCall.getStatus();
							   if (callStat==Call.Status.INPROGRESS || callStat==Call.Status.EARLYMEDIA) 
							   {	   
								
								   if (fixDigit<=11)
								   {	   
									 curCall.send(Call.DTMF.values()[fixDigit]);
								     log.debug("DTMF:"+fixDigit+" sent to skype");
								   }						  
							    }   
						   }
						   catch (Exception e)
						   {
							   if (callStat==Call.Status.EARLYMEDIA)
								     log.error("Error sending dtmf: "+fixDigit+" to skype during early media state");
							   else
							     log.error("Error sending dtmf: "+fixDigit+" to skype: ",e);
						   }
					   }
					   if (digitDelay>0)
						   try {Thread.sleep(digitDelay);} catch (Exception e) {}
				   }
				}
		}
		catch(Exception e)
		{}
	}
	
	private synchronized int getNextDtmfDigit() throws InterruptedException
	{
		while (dtmfQueueSize==0 && !stopSender)
			wait(1000); // just in case we miss a notify for some reason
		
		if (stopSender)
			return -1;
		
	    int digit=dtmfQueue.get(0);
	    dtmfQueue.remove(0);
	    dtmfQueueSize=dtmfQueue.size();

	    return digit;
	}
	
	public synchronized void queueDtmfDigit(int digit)
	{
		
	   dtmfQueue.add(digit);
	   dtmfQueueSize=dtmfQueue.size();

	   if (dtmfQueueSize==1)
		   notify();
	}


}
